# -*- coding: utf8 -*-
from __future__ import unicode_literals

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimString,
    Segment,
)


class VariantSegmentH(HeaderSegment):
    sender_id = HprimMultiple(
        5, system_name=HprimString(1), instrument_name=HprimString(2)
    )
    processing_id = HprimString(12, default="P")
    date = HprimDate(14)
    patients = HprimForeign("P", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class VariantSegmentP(Segment):
    patient_id = HprimString(3)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    mother_maiden_name = HprimString(7)
    birthday = HprimDate(8, fmt_date="%Y%m%d")
    # M/F/U
    sex = HprimString(9)
    physician_id = HprimString(14)
    orders = HprimForeign("O", "patient")


class VariantSegmentO(Segment):

    # numero de tube ou lot
    # Les deux niveaux LC et HC présent sur les deux premiers caractères
    # de ce champ permettent l'identification du QC
    sample_id = HprimString((3, 1))

    # numero de rack du tube
    vial = HprimString((3, 2))

    # position du tube sur le rack
    replicate = HprimString((3, 3))
    universal_test_id = HprimList(5, basefield=HprimMultiple(0, test_id=HprimString(4)))
    # S : Stat
    # R : Routine
    priority = HprimString(6)
    instrument_id = HprimString(25)
    results = HprimForeign("R", "order")


class VariantSegmentC(Segment):
    comment_source = HprimString(3)
    comment_text = HprimString(4)
    comment_typ = HprimString(5)


class VariantSegmentR(Segment):
    universal_test_id = HprimMultiple(
        3, peak_name=HprimString(4), component=HprimString(5)
    )
    value = HprimString(4)
    units = HprimString(5)

    # Not supported dans la doc
    # mais en réalité c'est renseigné alors que date_completed ne l'est pas...
    date_started = HprimDate(12)
    date_completed = HprimDate(13)
    instrument_identification = HprimString(14)


class VariantSegmentL(Segment):
    code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {"code": "variant", "label": "Variant / D100"}
    SEGMENT_CLASS_MAP = {
        "H": VariantSegmentH,
        "P": VariantSegmentP,
        "O": VariantSegmentO,
        "C": VariantSegmentC,
        "R": VariantSegmentR,
        "L": VariantSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"
    OPTIONS = [
        AstmMessageV2.BooleanOption(
            "variant_send_priority",
            label="Envoyer la priorité avec la demande dans le champ O.6",
            default=False,
        )
    ]

    def __init__(self, **kwargs):
        self.send_priority = kwargs.pop("variant_send_priority", False)
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = VariantSegmentH(type_segment="H", **kwargs)
        message.end = VariantSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        message = self.create(date=context["now"])
        # ignorer les suppressions car non gérées par le Variant
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE},
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
        ):
            seg_p = message.patients.add(
                patient_id=patient.patient_id,
                last_name=patient.last_name,
                first_name=patient.first_name,
                birthday=patient.birthdate,
                sex=patient.sex,
            )
            for physician in patient.physicians:
                if physician.type == "prescriber":
                    seg_p.physician_id = physician.code
                    break
            for sample, action, priority in patient.samples:
                seg_o = seg_p.orders.add(sample_id=sample.sample_id)
                if self.send_priority:
                    seg_o.priority = priority
                for order in sample.orders:
                    test = seg_o.universal_test_id.add()
                    test.test_id = order.code

        if message.patients:
            return message

    def decode_message(self, message):
        worklist = Worklist(context="result")
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                if seg_o.sample_id[0:2] in ("QC", "HC", "LC"):
                    sample = worklist.get_or_create(
                        type="control",
                        sample_id=seg_o.sample_id,
                        control_lot=seg_o.sample_id,
                    )
                else:
                    sample = worklist.get_or_create(
                        type="patient", sample_id=seg_o.sample_id
                    )
                sample.rack = "%s-%s" % (seg_o.vial, seg_o.replicate)
                for idx, seg_c in enumerate(seg_o.commentaires, 1):
                    sample.results.add(
                        code="COMMENT_%s_%s"
                        % (seg_o.universal_test_id[0].test_id, idx),
                        value=seg_c.comment_text,
                        date=message.date,
                    )
                for seg_r in seg_o.results:
                    sample.results.add(
                        code="%s%s"
                        % (
                            seg_r.universal_test_id.peak_name,
                            seg_r.universal_test_id.component,
                        ),
                        value=seg_r.value,
                        date=seg_r.date_completed or seg_r.date_started,
                        instrument=message.sender_id.instrument_name,
                    )
        return worklist
