# -*- coding: utf8 -*-

from __future__ import unicode_literals

import datetime
import json
from collections import OrderedDict

import jsonschema
import jsonschema_model
import six

import driverng.utils as utils

ORDER_SCHEMA = {
    "name": "Order",
    "type": "object",
    "properties": {
        "code": {"title": "Code", "type": "string"},
        "label": {"title": "Label", "type": "string"},
        "codification": {
            "title": "codification (L=Local LN=LOINC E=Euclidien)",
            "type": "string",
            "enum": ["", "L", "LN", "E"],
        },
        "alt_code": {"title": "Code alternatif", "type": "string"},
        "alt_label": {"title": "Libelle alternatif", "type": "string"},
        "alt_codification": {
            "title": "codification alternative (L=Local LN=LOINC E=Euclidien)",
            "type": "string",
            "enum": ["", "L", "LN", "E"],
        },
        "priority": {"title": "Priority", "type": "string"},
        "dilution": {"title": "Dilution", "type": "string"},
        "analyseur_repasse": {"title": "Analyseur demandé", "type": "string"},
        "action": {"title": "Action", "type": "string", "enum": ["new", "add", "del"]},
        "rerun": {"title": "rerun", "type": "boolean"},
        "dem_auto_test_status": {"title": "Status dem_auto_test", "type": "integer"},
        "nature_color": {"title": "Couleur de la nature", "type": "string"},
        "ordre": {"type": "number"},
        "order_id": {"type": "string"},
        "zone_groupement_test": {"title": "Zone groupement du test", "type": "string"},
        "loinc_code": {"title": "Code LOINC", "type": "string"},
        "accueil": {"title": "SIL Résult", "type": "boolean"},
    },
    "required": ["code", "action"],
    "additionalProperties": False,
}

LOT_SCHEMA = {
    "name": "Lot",
    "type": "object",
    "properties": {
        "type": {
            "title": "Type",
            "type": "string",
            "enum": ["reagent", "calibration", "control"],
        },
        "lot": {"title": "N°", "type": "string"},
        "name": {"title": "Name", "type": "string"},
        "expiration": {"title": "Expiration", "type": "date-time"},
    },
    "required": ["type", "lot"],
    "additionalProperties": False,
}

RESULT_SCHEMA = {
    "name": "Result",
    "type": "object",
    "properties": {
        "code": {"title": "Code", "type": "string"},
        "label": {"title": "Label", "type": "string"},
        "codification": {
            "title": "codification (L=Local LN=LOINC E=Euclidien)",
            "type": "string",
            "enum": ["", "L", "LN", "E"],
        },
        # Attention sub_code ce n'est pas la meme chose, c'est utilisé
        # dans des scénarios différents pour les résultats de bactériologie.
        "alt_code": {"title": "Code alternatif", "type": "string"},
        "alt_label": {"title": "Libelle alternatif", "type": "string"},
        "alt_codification": {
            "title": "codification alternatif (L=Local LN=LOINC E=Euclidien)",
            "type": "string",
            "enum": ["", "L", "LN", "E"],
        },
        "type": {"title": "Type", "type": "string"},
        "sub_code": {"title": "Sub code", "type": "string"},
        "value": {"title": "Value", "type": ["string", "array", "object"]},
        "atb": {"title": "Antibiogramme ATB", "type": "array"},
        # reserved for antibiogram results sent by clarilab
        "antibiogram": {
            "title": "Antibiogram",
            "name": "Antibiogram",
            "type": "array",
            "items": {
                "title": "Antibiotic",
                "name": "Antibiotic",
                "type": "object",
                "required": ["code"],
                "properties": {
                    "code": {"title": "Code", "type": "string"},
                    "label": {"title": "Label", "type": "string"},
                    "codification": {"title": "Codification", "type": "string"},
                    "value": {"title": "Value", "type": ["string", "array", "object"]},
                    "sensibility": {"title": "Sensibility", "type": "string"},
                    "units": {"title": "Units", "type": "string"},
                    "diameter": {"title": "Diameter", "type": "string"},
                    "cutoff": {"title": "CUT-OFF", "type": "string"},
                },
            },
        },
        "phenotypes": {"title": "Phenotypes", "type": "array"},
        "value_label": {"title": "Value labels", "type": ["string", "array", "object"]},
        "chapter": {
            "title": "Chapter",
            "name": "Chapitre",
            "type": "object",
            "properties": {
                "code": {"title": "Chapter code", "type": "string"},
                "label": {"title": "Chapter label", "type": "string"},
                "loinc_code": {"title": "Chapter LOINC code", "type": "string"},
                "position": {"title": "Chapter position (rang)", "type": "string"},
            },
        },
        "battery": {
            "title": "Battery (Analyse)",
            "name": "Analyse",
            "type": "object",
            "properties": {
                "code": {"title": "Battery code", "type": "string"},
                "chapter_position": {
                    "title": "Battery position (rang)",
                    "type": "string",
                },
                "position": {"title": "Battery position (rang)", "type": "string"},
            },
        },
        "resu_json": {
            "title": "Resu json",
            "type": "object",
            "properties": {
                "germes": {"title": "Germes", "type": "array"},
                "resultat": {"title": "Resultat", "type": "string"},
                "type_resultat": {"title": "Type resultat", "type": "string"},
            },
        },
        "graph": {"type": "object"},
        # Date d'intégration d'un résultat accueil pour l'envoi avec la demande
        # Elle correspond à la date RESULTAT.date
        # Dans le traitement des trames, il s'agit de la date du résultat
        # présent dans la trame
        "date": {"title": "Date", "type": "date-time"},
        # La date de l'instrument sert UNIQUEMENT à envoyer les résultats accueils
        # provenant d'un automate avec la demande. Il s'agit de la date brute de l'instrument
        # présente dans RESU_AUTO.date
        "date_instrument": {"title": "Date instrument", "type": "date-time"},
        "anterior": {"title": "Anterior", "type": "boolean"},
        "request_id": {"title": "Anterior request ID", "type": "string"},
        "request_date": {"title": "Anterior request date", "type": "date-time"},
        "normality": {"title": "Interpratation Normality", "type": "string"},
        "unit": {"title": "Unit", "type": "string"},
        "unit_label": {"title": "Unit Label", "type": "string"},
        "unit_codification": {
            "title": "Unit codification (L=Local LN=LOINC)",
            "type": "string",
            "enum": ["", "L", "LN"],
        },
        "alt_unit": {"title": "Alternative unit", "type": "string"},
        "alt_unit_label": {"title": "Alternative unit Label", "type": "string"},
        "alt_unit_codification": {
            "title": "Alternative unit codification (L=Local LN=LOINC)",
            "type": "string",
            "enum": ["", "L", "LN"],
        },
        "flags": {
            "title": "Flags",
            "type": "array",
            "items": {"title": "Flag", "type": "string"},
        },
        "indic_norm": {
            "title": "Indication normalité",
            "type": "array",
            "items": {"title": "Flag", "type": "string"},
        },
        "comments": {
            "title": "Comments",
            "type": "array",
            "items": {"title": "Comment", "type": "string"},
        },
        "instrument": {"title": "Instrument", "type": "string"},
        "status": {"title": "Status", "type": "string"},
        "validator": {"title": "Validator", "type": "string"},
        "validator_code": {"title": "Validator code", "type": "string"},
        "operator": {"title": "Operator", "type": "string"},
        "dilution": {"title": "Dilution", "type": "string"},
        "date_plateau": {"title": "Date plateau", "type": "date-time"},
        "nom_antibiogramme": {"title": "Nom antibiogramme", "type": "string"},
        "ligne": {"title": "Ligne", "type": "string"},
        "reagent": {"title": "Reagent", "type": "string"},
        "lots": {"type": "array", "items": LOT_SCHEMA},
        # Valide pour MCA/HPRIM uniquement
        "resultat_controle": {"type": "string"},
        "order_id": {"type": "string"},
        "reference_ranges": {"type": "string"},
        "preanalytique": {"title": "Pré-anlytique", "type": "boolean"},
        "zero": {
            "title": "Zeros",
            "type": "array",
            "items": {"title": "Zero", "type": "string"},
        },
        "loinc1_code": {"title": "Code LOINC Unité 1", "type": "string"},
        "loinc1_label": {"title": "Libellé LOINC Unité 1", "type": "string"},
        "loinc1_codification": {
            "title": "Codification LOINC Unité 1",
            "type": "string",
        },
        "loinc2_code": {"title": "Code LOINC Unité 2", "type": "string"},
        "loinc2_label": {"title": "Libellé LOINC Unité 2", "type": "string"},
        "loinc2_codification": {
            "title": "Codification LOINC Unité 2",
            "type": "string",
        },
    },
    "required": ["code", "value"],
    "additionalProperties": False,
}

BILLING_SCHEMA = {
    "name": "Billing",
    "type": "object",
    "properties": {
        "lettre_cle": {"title": "Lettre cle", "type": "string"},
        "coefficient": {"title": "Coefficient", "type": "number"},
        "denombrement": {"title": "Denombrement", "type": "number"},
        "quantite": {"title": "Quantite", "type": "number"},
        "libelle_acte": {"title": "Libelle de l'acte", "type": "string"},
        "type_majoration": {"title": "Type de majoration", "type": "string"},
        "no_executant": {"title": "Numero executant", "type": "string"},
        "code_executant": {"title": "Code executant", "type": "string"},
        "nom_executant": {"title": "Nom executant", "type": "string"},
        "code_nomenclature": {"title": "Code de nomenclature", "type": "string"},
        "catalogue": {"title": "Catalogue", "type": "string"},
        "code_rmo": {"title": "Code RMO", "type": "string"},
    },
    "additionalProperties": False,
}

SAMPLE_SCHEMA = {
    "name": "Sample",
    "type": "object",
    "properties": {
        "sample_id": {"title": "Sample ID", "type": "string"},
        "type": {
            "title": "Sample type",
            "type": "string",
            "enum": ["patient", "control"],
        },
        "request_id": {"title": "Request ID", "type": "string"},
        "request_info": {"title": "Request Info", "type": "string"},
        "request_mode_hbdrive": {"title": "Request Mode Hbdrive", "type": "string"},
        "exe_sample_id": {"title": "Sample ID (executant)", "type": "string"},
        "exe_request_id": {"title": "Request ID (executant)", "type": "string"},
        "receiver_request_id": {"title": "Receiver Request ID", "type": "string"},
        "control_lot": {"title": "Lot ID", "type": "string"},
        "control_name": {"title": "Lot Name", "type": "string"},
        "control_level": {"title": "Lot Level", "type": "string"},
        "rack": {"title": "Rack", "type": "string"},
        "sample_color": {"title": "Sample color", "type": "string"},
        "nature": {
            "title": "Nature de prélèvement",
            "name": "Nature",
            "type": "object",
            "properties": {
                "code": {"title": "code nature", "type": "string"},
                "label": {"title": "libelle nature", "type": "string"},
                "codification": {
                    "title": "codification (L=Local LN=LOINC)",
                    "type": "string",
                    "enum": ["", "L", "LN"],
                },
                "additive": {
                    "title": "additif ajoute au prelevement",
                    "type": "string",
                },
                "number_of_tubes": {"title": "Number of tubes", "type": "integer"},
                "barcode": {"title": "Barcode", "type": "string"},
            },
        },
        "barcode_media": {"title": "Barcode media", "type": "array"},
        "collection_date": {"title": "Collection date", "type": "date-time"},
        "collection_end_date": {"title": "Collection end date", "type": "date-time"},
        "request_date": {"title": "Request date", "type": "date-time"},
        "admission_date": {"title": "Admission date", "type": "date-time"},
        "admission_end_date": {"title": "Admission end date", "type": "date-time"},
        "admission_status": {"title": "Admission status", "type": "string"},
        "services": {
            "title": "Services",
            "type": "array",
            "name": "Services",
            "items": {
                "title": "Service",
                "name": "Service",
                "type": "object",
                "required": ["code"],
                "properties": {
                    "code": {"title": "Code", "type": "string"},
                    "name": {"title": "Name", "type": "string"},
                    "phone": {"title": "Phone", "type": "string"},
                },
            },
        },
        "physicians": {
            "title": "Physicians",
            "name": "Physicians",
            "type": "array",
            "items": {
                "title": "Physician",
                "name": "Physician",
                "type": "object",
                "required": ["code"],
                "properties": {
                    "code": {"title": "Code", "type": "string"},
                    "type": {
                        "title": "Type",
                        "type": "string",
                        "enum": [
                            "attending",
                            "sampler",
                            "prescriber",  # referring
                            "admitting",
                            "consulting",
                            "family",
                            "referred",
                            "official",
                            "substitute",
                        ],
                    },
                    "last_name": {"title": "Last name", "type": "string"},
                    "first_name": {"title": "First name", "type": "string"},
                    "phone": {"title": "Phone", "type": "string"},
                    "codification": {"title": "Codification", "type": "string"},
                    "alias": {"title": "Alias", "type": "string"},
                    "title": {"title": "Title", "type": "string"},
                    "speciality": {"title": "Title", "type": "string"},
                    "action_code": {"title": "Action code", "type": "string"},
                    "further_names": {
                        "title": "Further Names",
                        "type": "array",
                        "items": {"title": "Further name", "type": "string"},
                    },
                    "code_type": {"title": "Code type", "type": "string"},
                    "oid": {"title": "OID", "type": "string"},
                    "begin_date": {"title": "Begin date", "type": "date-time"},
                    "end_date": {"title": "End date", "type": "date-time"},
                },
            },
        },
        "relations": {
            "title": "Relations",
            "name": "Relations",
            "type": "array",
            "items": {
                "title": "Relation",
                "name": "Relation",
                "type": "object",
                "properties": {
                    "last_name": {"title": "Last name", "type": "string"},
                    "used_last_name": {"title": "Used last name", "type": "string"},
                    "first_name": {"title": "First name", "type": "string"},
                    "used_first_name": {"title": "Used first name", "type": "string"},
                    "further_names": {
                        "title": "Further Names",
                        "type": "array",
                        "items": {"title": "Further name", "type": "string"},
                    },
                    "title": {"title": "Title", "type": "string"},
                    "identity_attribut": {"title": "Name Type Code", "type": "string"},
                    "relationship": {"title": "Relationship", "type": "string"},
                    "address": {"title": "Address", "type": "string"},
                    "other_designation": {
                        "title": "Other designation",
                        "type": "string",
                    },
                    "postal_code": {"title": "Postal code", "type": "string"},
                    "city": {"title": "City", "type": "string"},
                    "country": {"title": "Country", "type": "string"},
                    "address_type": {"title": "Type", "type": "string"},
                    "county_code": {"title": "County code", "type": "string"},
                    "phone": {"title": "Phone", "type": "string"},
                    "mobile": {"title": "Mobile", "type": "string"},
                    "email": {"title": "Email", "type": "string"},
                    "role": {"title": "Role", "type": "string"},
                    "identifier": {"title": "Identifier", "type": "string"},
                    "oid": {"title": "Authority", "type": "string"},
                },
            },
        },
        # LEGACY
        "service": {"type": "string"},
        "service_name": {"type": "string"},
        "physician_name": {"type": "string"},
        "physician_first_name": {"type": "string"},
        "medecin2": {"type": "string"},
        "request_comment_patient": {"type": "string"},
        "request_comment_obr": {"type": "string"},
        # END LEGACY
        "transcriptionist": {"title": "Transcriptionist", "type": "string"},
        "patient_class": {"title": "Classe de patient", "type": "string"},
        "location": {"title": "Location", "type": "string"},
        "location_name": {"title": "Location name", "type": "string"},
        "room": {"title": "Room", "type": "string"},
        "bed": {"title": "Bed", "type": "string"},
        "bed_status": {"title": "Bed Status", "type": "string"},
        "hospital": {"title": "Hospital", "type": "string"},
        "hospital_service": {"title": "Hospital Service", "type": "string"},
        "last_name": {"title": "Last name", "type": "string"},
        "used_last_name": {"title": "Used last name", "type": "string"},
        "further_names": {
            "title": "Further Names",
            "type": "array",
            "items": {"title": "Further name", "type": "string"},
        },
        "first_name": {"title": "First name", "type": "string"},
        "used_first_name": {"title": "Used first name", "type": "string"},
        "title": {"title": "Title", "type": "string"},
        "identity_attribut": {"title": "Name Type Code", "type": "string"},
        "birthdate": {"title": "Birth date", "type": "date-time"},
        "lunar_date": {"title": "Lunar Date", "type": "date-time"},
        "gestation_weeks": {"title": "Gestation Weeks", "type": "string"},
        "sms_consent": {"title": "SMS Consent", "type": "string"},
        "birthdate_modified": {"title": "Birth date modified", "type": "string"},
        "identity_obtaining_mode": {
            "title": "Identity obtaining mode",
            "type": "string",
        },
        "insi_query_date": {"title": "INSi query date", "type": "date-time"},
        "identity_document": {"title": "Identity document", "type": "string"},
        "identity_document_expiration_date": {
            "title": "Identity document expiration date",
            "type": "date-time",
        },
        "birthplace": {"title": "Birth place", "type": "string"},
        "birthplace_code": {"title": "Birth place code", "type": "string"},
        "deathdate": {"title": "Date of death", "type": "date-time"},
        "birth_row": {"title": "Birth row", "type": "string"},
        "maiden_name": {"title": "Maiden name", "type": "string"},
        "mother_identity": {
            "title": "Mother Identity",
            "type": "object",
            "properties": {
                "last_name": {"title": "Last Name", "type": "string"},
                "first_name": {"title": "First Name", "type": "string"},
                "further_names": {
                    "title": "Further Names",
                    "type": "array",
                    "items": {"title": "Further name", "type": "string"},
                },
                "prefix": {"title": "Prefix", "type": "string"},
                "name_type_code": {"title": "Name Type Code", "type": "string"},
                "identifier": {"title": "Identifier", "type": "string"},
                "assigning_authority": {
                    "title": "Assigning Authority",
                    "type": "string",
                },
                "oid": {"title": "OID", "type": "string"},
            },
        },
        "postal_code": {"title": "Postal code", "type": "string"},
        "county_code": {"title": "Code de la commune", "type": "string"},
        "city": {"title": "City", "type": "string"},
        "country": {"title": "Country", "type": "string"},
        "address": {"title": "Address", "type": "string"},
        "address_effective_date": {
            "title": "Date de début de validité de l'adresse",
            "type": "date-time",
        },
        "address_expiration_date": {
            "title": "Date de fin de validité de l'adresse",
            "type": "date-time",
        },
        "lab": {"title": "Laboratory", "type": "string"},
        "sex": {"title": "Sex", "type": "string"},
        "patient_id": {"title": "Patient ID", "type": "string"},
        "ins_number": {"title": "Matricule INS", "type": "string"},
        "old_ins_number": {"title": "Matricule INS à supprimer", "type": "string"},
        "old_ipp": {"title": "IPP à supprimer", "type": "string"},
        "oid": {"title": "OID", "type": "string"},
        "identity_status": {"title": "Identity status", "type": "string"},
        "lab_patient_id": {"title": "Laboratory patient ID", "type": "string"},
        "alternate_patient_id": {"title": "Alternate patient ID", "type": "string"},
        "specimen": {"title": "Specimen code", "type": "string"},
        "diagnostic": {"title": "Diagnostic", "type": "string"},
        "clinical_information": {"title": "Clinical Information", "type": "string"},
        "height": {"title": "Height", "type": "string"},
        "weight": {"title": "Weight", "type": "string"},
        "phone": {"title": "Phone", "type": "string"},
        "mobile": {"title": "Mobile", "type": "string"},
        "email": {"title": "Email", "type": "string"},
        "pro_phone": {"title": "Pro Phone", "type": "string"},
        "pro_mobile": {"title": "Pro Mobile", "type": "string"},
        "pro_email": {"title": "Pro Email", "type": "string"},
        "medications": {
            "title": "Medications",
            "type": "array",
            "items": {"type": "string"},
        },
        "priority": {"title": "Priority", "type": "string"},
        "physiologie": {"title": "Physiologie", "type": "string"},
        "patient_comment1": {"title": "Patient comment1", "type": "string"},
        "patient_comment2": {"title": "Patient comment2", "type": "string"},
        "media_code": {"title": "Media code", "type": "string"},
        "media_suffix_code": {"title": "Media suffix code", "type": "string"},
        "barcode_secondary_media": {
            "title": "Barcode secondary media",
            "type": "array",
            "items": {"type": "string"},
        },
        "bruker_projet": {"title": "Bruker projet", "type": "string"},
        "bruker_projet_name": {"title": "Bruker Name projet", "type": "string"},
        "bruker_type_echantillon": {
            "title": "Bruker type echantillon",
            "type": "string",  # TODO: Maybe int
        },
        "bruker_position": {"title": "Bruker position", "type": "string"},
        "bruker_positions": {
            "title": "Bruker list position with sample Id",
            "type": "array",
        },
        "rack_dms_tracking_serotheque": {
            "title": "Rack Dms Tracking serotheque",
            "type": "string",
        },
        "sequence_scanlab": {"title": "Sequence Scanlab", "type": "string"},
        "comments": {
            "title": "Comments",
            "type": "array",
            "items": {
                "title": "Comment",
                "type": "object",
                "name": "Comment",
                "properties": {
                    "text": {"title": "Text", "type": "string"},
                    "type": {
                        "title": "Type",
                        "type": "string",
                        "enum": ["patient", "sample"],
                    },
                },
            },
        },
        "flags": {
            "title": "Flags",
            "type": "array",
            "items": {"title": "Flag", "type": "string"},
        },
        "orders": {"type": "array", "items": ORDER_SCHEMA},
        "results": {"type": "array", "items": RESULT_SCHEMA},
        "billings": {"type": "array", "items": BILLING_SCHEMA},
        "info_serotheque": {
            "title": "Serotheque",
            "type": "object",
            "properties": {
                "code": {"title": "code serotheque", "type": "string"},
                "portoir": {"title": "code portoir", "type": "string"},
                "sample_id": {"title": "Sample ID", "type": "string"},
                "numero": {"title": "numero serotheque", "type": "string"},
                "destination": {"title": "destination serotheque", "type": "string"},
                "position": {"title": "position serotheque", "type": "string"},
                "etiquette": {"title": "etiquette serotheque", "type": "string"},
            },
        },
        "language": {"title": "Language", "type": "string"},
        "marital_status": {"title": "Marital status", "type": "string"},
        "birth_order": {"title": "Birth order", "type": "string"},
        "citizenships": {
            "title": "Citizenships",
            "type": "array",
            "items": {"title": "Citizenship", "type": "string"},
        },
        "identity_update_date": {"title": "Identity update date", "type": "date-time"},
        "living_arrangement": {"title": "Living arrangement", "type": "string"},
        "data_protection": {"title": "Data protection", "type": "string"},
        "event": {
            "title": "Event",
            "type": "object",
            "properties": {
                "type_code": {"title": "Type Code", "type": "string"},
                "type_label": {"title": "type_label", "type": "string"},
                "recorded_date": {"title": "Recorded Date", "type": "date-time"},
                "occured_date": {"title": "Occured Date", "type": "date-time"},
            },
        },
        # specific elements for PAM ADT ITI-31
        "account_number": {
            "title": "Account Number",
            "type": "object",
            "properties": {
                "identifier": {"title": "Identifier", "type": "string"},
                "assigning_authority": {
                    "title": "Assigning Authority",
                    "type": "string",
                },
                "oid": {"title": "OID", "type": "string"},
            },
        },
        "visit_infos": {
            "title": "Visit Infos",
            "type": "object",
            "properties": {
                "admission_type": {"title": "Admission Type", "type": "string"},
                "preadmit_number": {"title": "Preadmit Number", "type": "string"},
                "preadmit_number_oid": {
                    "title": "Preadmit Number OID",
                    "type": "string",
                },
                "admit_source": {"title": "Admit Source", "type": "string"},
                "vip_indicator": {"title": "VIP Indicator", "type": "string"},
                "visit_number": {"title": "Visit Number", "type": "string"},
                "visit_number_oid": {"title": "Visit Number OID", "type": "string"},
                "financial_class": {"title": "Financial Class", "type": "string"},
                "charge_price_indicator": {
                    "title": "Charge Price Indicator",
                    "type": "string",
                },
                "courtesy_code": {"title": "Courtesy Code", "type": "string"},
                "discharge_disposition": {
                    "title": "Discharge Disposition",
                    "type": "string",
                },
                "discharge_to_location": {
                    "title": "Discharge To Location",
                    "type": "string",
                },
                "account_status": {"title": "Account Status", "type": "string"},
                "accommodation_code": {"title": "Accommodation Code", "type": "string"},
                "visit_user_code": {"title": "Visit User Code", "type": "string"},
                "expected_admit_date": {
                    "title": "Expected Admit Date",
                    "type": "date-time",
                },
                "expected_discharge_date": {
                    "title": "Expected Discharge Date",
                    "type": "date-time",
                },
                "visit_description": {"title": "Visit Description", "type": "string"},
                "special_program_code": {
                    "title": "Special Program Code",
                    "type": "string",
                },
                "visit_protection_indicator": {
                    "title": "Visit Protection Indicator",
                    "type": "string",
                },
                "patient_charge_ajustment_code": {
                    "title": "Patient Charge Ajustment_code",
                    "type": "string",
                },
                "mode_of_arrival_code": {
                    "title": "Mode Of Arrival Code",
                    "type": "string",
                },
                "expected_loa_return_date": {
                    "title": "Expected LOA Return Date",
                    "type": "date-time",
                },
                "admit_date": {"title": "Admit Date", "type": "date-time"},
                "discharge_date": {"title": "Discharge Date", "type": "date-time"},
            },
        },
        "movements": {
            "title": "Movements",
            "type": "object",
            "properties": {
                "identifier": {"title": "Identifier", "type": "string"},
                "assigning_authority": {"title": "Namespace ID", "type": "string"},
                "oid": {"title": "Universal ID", "type": "string"},
                "start_date": {"title": "Start Date", "type": "date-time"},
                "action": {"title": "Action", "type": "string"},
                "historic_indicator": {"title": "Historic Indicator", "type": "string"},
                "original_trigger": {"title": "Original Trigger", "type": "string"},
                "ward_name": {"title": "Ward Name", "type": "string"},
                "ward_assigning_authority": {
                    "title": "Ward Assigning Authority",
                    "type": "string",
                },
                "ward_organization_id": {
                    "title": "Ward Organization ID",
                    "type": "string",
                },
                "nursing_ward_name": {"title": "Nursing Ward Name", "type": "string"},
                "nursing_ward_assigning_authority": {
                    "title": "Nursing Ward Assigning Authority",
                    "type": "string",
                },
                "nursing_ward_organization_id": {
                    "title": "Nursing Ward Organization ID",
                    "type": "string",
                },
                "scope": {"title": "Scope", "type": "string"},
            },
        },
        "dmp": {
            "title": "DMP",
            "type": "object",
            "properties": {
                "dmp_status": {"title": "DMP Status", "type": "string"},
                "collecting_date": {"title": "Collecting Date", "type": "date-time"},
                "closing_date": {"title": "Closing Date", "type": "date-time"},
                "access_authorization": {
                    "title": "Access Authorization",
                    "type": "string",
                },
                "access_authorization_collecting_date": {
                    "title": "Access Authorization Collecting Date",
                    "type": "date-time",
                },
                "breakage_mode_opposition": {
                    "title": "Breakage Mode Opposition",
                    "type": "string",
                },
                "control_center_mode_opposition": {
                    "title": "Control Center Mode Opposition",
                    "type": "string",
                },
                "opposition_collecting_date": {
                    "title": "Opposition Collecting Date",
                    "type": "date-time",
                },
            },
        },
        "professional_infos": {
            "title": "Professional Infos",
            "type": "object",
            "properties": {
                "socio_professional_activity": {
                    "title": "Socio Professional Activity",
                    "type": "string",
                },
                "socio_professional_category": {
                    "title": "Socio Professional Category",
                    "type": "string",
                },
            },
        },
        "accident_infos": {
            "title": "Accident Infos",
            "type": "object",
            "properties": {
                "accident_date": {"title": "Accident Date", "type": "date-time"},
                "accident_code": {"title": "Accident Code", "type": "string"},
                "accident_text": {"title": "Accident Text", "type": "string"},
            },
        },
        "pmsi": {
            "title": "PMSI",
            "type": "object",
            "properties": {
                "entry_mode": {"title": "Entry Mode", "type": "string"},
                "exit_mode": {"title": "Exit Mode", "type": "string"},
                "origin_mode": {"title": "Origin Mode", "type": "string"},
                "destination_mode": {"title": "Entry Mode", "type": "string"},
            },
        },
        "previous_facility": {
            "title": "Previous Facility",
            "type": "object",
            "properties": {
                "name": {"title": "Name", "type": "string"},
                "entry_date": {"title": "Entry Date", "type": "date-time"},
                "transport_mode": {"title": "Transport Mode", "type": "string"},
                "psy_start_date": {"title": "Psy Start Date", "type": "date-time"},
                "psy_end_date": {"title": "Psy End Date", "type": "date-time"},
                "ori_address": {"title": "ORI Address", "type": "string"},
                "ori_city": {"title": "ORI City", "type": "string"},
                "ori_postal_code": {"title": "ORI Postal Code", "type": "string"},
                "ori_country": {"title": "ORI Country", "type": "string"},
                "ori_county_code": {"title": "ORI County Code", "type": "string"},
                "dst_address": {"title": "DST Address", "type": "string"},
                "dst_city": {"title": "DST City", "type": "string"},
                "dst_postal_code": {"title": "DST Postal Code", "type": "string"},
                "dst_country": {"title": "DST Country", "type": "string"},
                "dst_county_code": {"title": "DST County Code", "type": "string"},
                "nda": {"title": "NDA", "type": "string"},
                "archive_number": {"title": "Archive Number", "type": "string"},
                "custom_exit_mode": {"title": "Custom Exit Mode", "type": "string"},
                "rimp_code": {"title": "RIM-P Code", "type": "string"},
                "rimp_label": {"title": "RIM-P Label", "type": "string"},
                "transport_support_code": {
                    "title": "Transport Support Code",
                    "type": "string",
                },
                "transport_support_label": {
                    "title": "Transport Support Label",
                    "type": "string",
                },
            },
        },
        "guarantor": {
            "title": "Guarantor",
            "type": "object",
            "properties": {
                "identifier": {"title": "Identifier", "type": "string"},
                "assigning_authority": {
                    "title": "Assigning Authority",
                    "type": "string",
                },
                "oid": {"title": "OID", "type": "string"},
                "last_name": {"title": "Last Name", "type": "string"},
                "first_name": {"title": "First Name", "type": "string"},
                "further_names": {
                    "title": "Further Names",
                    "type": "array",
                    "items": {"title": "Further name", "type": "string"},
                },
                "prefix": {"title": "Prefix", "type": "string"},
                "name_type_code": {"title": "Name Type Code", "type": "string"},
                "address": {"title": "Address", "type": "string"},
                "city": {"title": "City", "type": "string"},
                "postal_code": {"title": "Postal Code", "type": "string"},
                "country": {"title": "Country", "type": "string"},
                "county_code": {"title": "County Code", "type": "string"},
                "address_type": {"title": "Address Type", "type": "string"},
                "address_effective_date": {
                    "title": "Address Effective Date",
                    "type": "date-time",
                },
                "address_expiration_date": {
                    "title": "Address Expiration Date",
                    "type": "date-time",
                },
                "phone": {"title": "Phone", "type": "string"},
                "mobile": {"title": "Mobile", "type": "string"},
                "email": {"title": "Email", "type": "string"},
                "pro_phone": {"title": "Pro Phone", "type": "string"},
                "pro_mobile": {"title": "Pro Mobile", "type": "string"},
                "pro_email": {"title": "Email", "type": "string"},
                "type": {"title": "Type", "type": "string"},
                "relationship": {"title": "Relationship", "type": "string"},
                "ssn": {"title": "SSN", "type": "string"},
                "date_begin": {"title": "Date Begin", "type": "date-time"},
                "date_end": {"title": "Date End", "type": "date-time"},
                "priority": {"title": "Priority", "type": "string"},
                "employer_identifier": {
                    "title": "Employer Identifier",
                    "type": "string",
                },
                "employer_assigning_authority": {
                    "title": "Employer Assigning Authority",
                    "type": "string",
                },
                "employer_oid": {"title": "Employer OID", "type": "string"},
                "employer_last_name": {"title": "Employer Last Name", "type": "string"},
                "employer_first_name": {
                    "title": "Employer First Name",
                    "type": "string",
                },
                "employer_further_names": {
                    "title": "Employer Further Names",
                    "type": "array",
                    "items": {"title": "Employer Further name", "type": "string"},
                },
                "employer_prefix": {"title": "Employer Prefix", "type": "string"},
                "employer_name_type_code": {
                    "title": "Employer Name Type Code",
                    "type": "string",
                },
                "employer_address": {"title": "Employer Address", "type": "string"},
                "employer_city": {"title": "Employer City", "type": "string"},
                "employer_postal_code": {
                    "title": "Employer Postal Code",
                    "type": "string",
                },
                "employer_country": {"title": "Employer Country", "type": "string"},
                "employer_county_code": {
                    "title": "Employer County Code",
                    "type": "string",
                },
                "employer_address_type": {
                    "title": "Employer Address Type",
                    "type": "string",
                },
                "employer_address_effective_date": {
                    "title": "Employer Address Effective Date",
                    "type": "date-time",
                },
                "employer_address_expiration_date": {
                    "title": "Employer Address Expiration Date",
                    "type": "date-time",
                },
                "employer_phone": {"title": "Employer Phone", "type": "string"},
                "employer_mobile": {"title": "Employer Mobile", "type": "string"},
                "employer_email": {"title": "Employer Email", "type": "string"},
                "employer_organization_name": {
                    "title": "Employer Organization Name",
                    "type": "string",
                },
                "employer_organization_assigning_authority": {
                    "title": "Employer Organization Assigning Authority",
                    "type": "string",
                },
                "employer_organization_identifier": {
                    "title": "Employer Organization Identifier",
                    "type": "string",
                },
                "employee_id_number": {"title": "Employee ID Number", "type": "string"},
                "employement_status": {"title": "Employement Status", "type": "string"},
                "organization_name": {"title": "Organization Name", "type": "string"},
                "billing_hold_flag": {"title": "Billing Hold Flag", "type": "string"},
                "credit_rating_code": {"title": "Credit Rating Code", "type": "string"},
                "death_date": {"title": "Death Date", "type": "date-time"},
                "death_flag": {"title": "Death Flag", "type": "string"},
                "charge_ajustement_code": {
                    "title": "Charge Ajustement Code",
                    "type": "string",
                },
                "household_annual_income": {
                    "title": "Household Annual Income",
                    "type": "string",
                },
                "household_size": {"title": "Household Size", "type": "string"},
                "marital_status_code": {
                    "title": "Marital Status Code",
                    "type": "string",
                },
                "hire_effective_date": {
                    "title": "Hire Effective Date",
                    "type": "date-time",
                },
                "employement_stop_date": {
                    "title": "Employement Stop Date",
                    "type": "date-time",
                },
                "primary_language": {"title": "Primary Language", "type": "string"},
                "living_arrangement": {"title": "Living Arrangement", "type": "string"},
                "nationality": {"title": "Nationality", "type": "string"},
                "contact_last_name": {"title": "Contact Last Name", "type": "string"},
                "contact_first_name": {"title": "Contact First Name", "type": "string"},
                "contact_further_names": {
                    "title": "Contact Further Names",
                    "type": "array",
                    "items": {"title": "Contact Further name", "type": "string"},
                },
                "contact_prefix": {"title": "Contact Prefix", "type": "string"},
                "contact_name_type_code": {
                    "title": "Contact Name Type Code",
                    "type": "string",
                },
                "contact_phone": {"title": "Contact Phone", "type": "string"},
                "contact_mobile": {"title": "Contact Mobile", "type": "string"},
                "contact_email": {"title": "Contact Email", "type": "string"},
                "contact_reason": {"title": "Contact Reason", "type": "string"},
                "contact_relationship": {
                    "title": "Contact Relationship",
                    "type": "string",
                },
                "job_title": {"title": "Job Title", "type": "string"},
                "job_code": {"title": "Job Code", "type": "string"},
                "job_status": {"title": "Job Status", "type": "string"},
                "financial_class": {"title": "Financial Class", "type": "string"},
                "birth_place": {"title": "Birth Place", "type": "string"},
                "vip_indicator": {"title": "VIP Indicator", "type": "string"},
            },
        },
        "insurance": {
            "title": "Insurance",
            "type": "object",
            "properties": {
                "plan_id": {"title": "Plan ID", "type": "string"},
                "company_identifier": {"title": "Company Identifier", "type": "string"},
                "company_assigning_authority": {
                    "title": "Company Assigning Authority",
                    "type": "string",
                },
                "company_oid": {"title": "Company OID", "type": "string"},
                "plan_effective_date": {
                    "title": "Plan Effective Date",
                    "type": "date-time",
                },
                "plan_expiration_date": {
                    "title": "Plan Expiration Date",
                    "type": "date-time",
                },
                "authorization_name": {"title": "Authorization Name", "type": "string"},
                "authorization_date": {
                    "title": "Authorization Date",
                    "type": "date-time",
                },
                "exemption_from_copayment": {
                    "title": "Exemption From Copayment",
                    "type": "string",
                },
                "insured_last_name": {"title": "Insured Last Name", "type": "string"},
                "insured_first_name": {"title": "Insured First Name", "type": "string"},
                "insured_further_names": {
                    "title": "Insured Further Names",
                    "type": "array",
                    "items": {"title": "Insured Further name", "type": "string"},
                },
                "insured_prefix": {"title": "Insured Prefix", "type": "string"},
                "insured_name_type_code": {
                    "title": "Insured Name Type Code",
                    "type": "string",
                },
                "beneficiary_status": {"title": "Beneficiary Status", "type": "string"},
                "insured_address": {"title": "Insured Address", "type": "string"},
                "insured_city": {"title": "Insured City", "type": "string"},
                "insured_postal_code": {
                    "title": "Insured Postal Code",
                    "type": "string",
                },
                "insured_country": {"title": "Insured Country", "type": "string"},
                "insured_address_type": {
                    "title": "Insured Address Type",
                    "type": "string",
                },
                "insured_county_code": {
                    "title": "Insured County Code",
                    "type": "string",
                },
                "insured_address_effective_date": {
                    "title": "Insured Address Effective Date",
                    "type": "date-time",
                },
                "insured_address_expiration_date": {
                    "title": "Insured Address Expiration Date",
                    "type": "date-time",
                },
                "third_party_payer": {"title": "Third Party Payer", "type": "string"},
                "insurance_nature": {"title": "Insurance Nature", "type": "string"},
                "company_plan_code": {"title": "Company Plan Code", "type": "string"},
                "event_number": {"title": "Event Number", "type": "string"},
                "justificative_nature": {
                    "title": "Justificative Nature",
                    "type": "string",
                },
                "insured_identifier": {"title": "Insured Identifier", "type": "string"},
                "insured_assigning_authority": {
                    "title": "Insured Assigning Authority",
                    "type": "string",
                },
                "insured_oid": {"title": "Insured OID", "type": "string"},
                "room_coverage_type": {"title": "Room Coverage Type", "type": "string"},
                "room_coverage_amount": {
                    "title": "Room Coverage Amount",
                    "type": "string",
                },
                "insured_phone": {"title": "Insured Phone", "type": "string"},
                "insured_mobile": {"title": "Insured Mobile", "type": "string"},
                "insured_email": {"title": "Insured Email", "type": "string"},
                "money_or_percentage_indicator": {
                    "title": "Money Or Percentage Indicator",
                    "type": "string",
                },
                "money_or_percentage_quantity": {
                    "title": "Money Or Percentage Quantity",
                    "type": "string",
                },
            },
        },
    },
    "required": ["sample_id", "type"],
    "additionalProperties": False,
}

DICTIONARY_SCHEMA = {
    "name": "Dictionary",
    "type": "object",
    "properties": {
        "emetteur": {"title": "emetteur du dictionnaire", "type": "string"},
        "production": {"title": "message production ou mode test", "type": "boolean"},
        "unique_identifier": {
            "title": "identifiant unique du message",
            "type": "string",
        },
        "version": {"title": "version du catalogue (ne pas parser)", "type": "string"},
    },
    "entries": {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "pkey": {"title": "entrie pkey", "type": "string"},
                "code": {"title": "code examen", "type": "string"},
                "libelle": {"title": "libelle examen", "type": "string"},
                "codification": {
                    "title": "codification (L=Local LN=LOINC)",
                    "type": "string",
                    "enum": ["L", "LN"],
                },
                "sample_required": {"title": "echantillon requis", "type": "boolean"},
                "method": {"title": "technique", "type": "string"},
                "battery": {
                    "title": "examen atomique ou batterie d'examens",
                    "type": "boolean",
                },
                "turnaroundtime": {
                    "title": "delai de realisation en minutes",
                    "type": "integer",
                },
                "frequency": {
                    "title": "frequence de realisation (OM1-40)",
                    "type": "string",
                },
                "comments": {"title": "commentaires", "type": "string"},
                "price_hn_eur_cents": {
                    "title": "prix hors nomenclature (en centime d'euros)",
                    "type": "integer",
                },
                "price_fixed": {"title": "prix fixe", "type": "boolean"},
                "entente_prealable": {
                    "title": "entente prealable necessaire",
                    "type": "boolean",
                },
                "consentement": {
                    "title": "consentement patient necessaire",
                    "type": "boolean",
                },
                "analyse_declenchee": {
                    "title": "code analyse à declencher obligatoirement",
                    "type": "string",
                },
                "nabms": {
                    "title": "codes nabm",
                    "type": "array",
                    "items": {"title": "code nabm", "type": "string"},
                },
                "url_doc": {"title": "url documentaire", "type": "string"},
                "conditions_tarifaires": {
                    "title": "conditions tarifaires",
                    "type": "string",
                },
                "samples": {
                    "title": "echantillons",
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "container_description": {
                                "title": "description du contenant",
                                "type": "string",
                            },
                            "container_volume": {
                                "title": "volume du contenant",
                                "type": "integer",
                            },
                            "container_unit": {
                                "title": "unite de volume du contenant",
                                "type": "string",
                            },
                            "nature_code": {"title": "code nature", "type": "string"},
                            "nature_libelle": {
                                "title": "libelle nature",
                                "type": "string",
                            },
                            "alt_nature_code": {
                                "title": "code nature (alternatif)",
                                "type": "string",
                            },
                            "alt_nature_libelle": {
                                "title": "label nature (alternatif)",
                                "type": "string",
                            },
                            "additive_code": {
                                "title": "code additif",
                                "type": "string",
                            },
                            "additive": {
                                "title": "additif ajoute au prelevement",
                                "type": "string",
                            },
                            "transport_conditions_code": {
                                "title": "conditions de transport (code)",
                                "type": "string",
                            },
                            "transport_conditions_libelle": {
                                "title": "conditions de transport (libelle)",
                                "type": "string",
                            },
                            "minimum_volume": {
                                "title": "volume minimum necessaire ([valeur] [unite])",
                                "type": "string",
                            },
                        },
                    },
                },
            },
            "required": ["code", "libelle"],
        },
    },
    "required": ["entries"],
    "additionalProperties": False,
}

PATIENT_ATTRIBUTES = (
    "last_name",
    "used_last_name",
    "first_name",
    "used_first_name",
    "further_names",
    "identity_attribut",
    "birthdate",
    "lunar_date",
    "gestation_weeks",
    "sms_consent",
    "birthdate_modified",
    "maiden_name",
    "mother_identity",
    "sex",
    "patient_id",
    "ins_number",
    "old_ins_number",
    "old_ipp",
    "oid",
    "identity_status",
    "lab",
    "lab_patient_id",
    "postal_code",
    "county_code",
    "address",
    "address_effective_date",
    "address_expiration_date",
    "city",
    "birthplace",
    "birthplace_code",
    "country",
    "collection_date",
    "services",
    "physicians",
    "relations",
    "patient_class",
    "location",
    "location_name",
    "bed",
    "bed_status",
    "room",
    "hospital",
    "hospital_service",
    "admission_date",
    "admission_end_date",
    "admission_status",
    "height",
    "weight",
    "title",
    "alternate_patient_id",
    "phone",
    "mobile",
    "email",
    "pro_phone",
    "pro_mobile",
    "pro_email",
    "birth_row",
    "deathdate",
    # Legacy
    "service",
    "service_name",
    "physician",
    "physician_name",
    "physician_first_name",
    "comments",
    "medications",
    "patient_comment1",
    "patient_comment2",
    "diagnostic",
    "language",
    "marital_status",
    "birth_order",
    "citizenship",
    "living_arrangement",
    "data_protection",
    "event",
    "account_number",
    "visit_infos",
    "movements",
    "dmp",
    "professional_infos",
    "accident_infos",
    "pmsi",
    "previous_facility",
    "guarantor",
    "insurance",
)

PATIENT_SCHEMA = {
    "name": "Patient",
    "type": "object",
    "properties": dict(
        [
            (k, v)
            for k, v in SAMPLE_SCHEMA["properties"].items()
            if k in PATIENT_ATTRIBUTES
        ]
    ),
    "required": ["patient_id"],
    "additionalProperties": True,
}


SCHEMA = {
    "name": "Worklist",
    "type": "object",
    "required": ["context"],
    "properties": {
        "samples": {"type": "array", "items": SAMPLE_SCHEMA},
        "dictionary": {"type": "object", "properties": DICTIONARY_SCHEMA},
        "context": {
            "title": "Context",
            "type": "string",
            "enum": ["order", "query", "result", "billing", "dictionary", "info"],
        },
        "origin": {"title": "Origin", "type": "string", "enum": ["bruker"]},
        "filename": {"title": "Filename", "type": "string"},
        "sender_id": {"title": "Sender ID", "type": "string"},
        "sender_name": {"title": "Sender Name", "type": "string"},
        "receiver_id": {"title": "Receiver ID", "type": "string"},
        "receiver_name": {"title": "Receiver Name", "type": "string"},
        "orig_data": {"type": "string"},
        "data": {"type": "string"},
        "traceback": {"type": "string"},
    },
    "additionalProperties": False,
}


Model = jsonschema_model.model_factory(SCHEMA)
Patient = jsonschema_model.model_factory(PATIENT_SCHEMA)
Sample = jsonschema_model.model_factory(SAMPLE_SCHEMA)
Order = jsonschema_model.model_factory(ORDER_SCHEMA)
Result = jsonschema_model.model_factory(RESULT_SCHEMA)
Lot = jsonschema_model.model_factory(LOT_SCHEMA)
Billing = jsonschema_model.model_factory(BILLING_SCHEMA)
Dictionary = jsonschema_model.model_factory(DICTIONARY_SCHEMA)


class Worklist(Model):
    """Format natif pour représenter un message

    La classe est basée sur un schema json et construite avec
    `jsonschema-model <https://github.com/philpep/jsonschema-model>`_

    Exemple de message de demandes::

        {
            "context": "order",
            "samples": [{
                "type": "patient",
                "sample_id": "123456789",
                "first_name": "ÉTIENNE",
                "last_name": "Bézout,
                "birthdate": datetime.datetime(1730, 3, 31),
                "specimen": "Serum",
                "orders": [
                    {"code": "A", "action": "add", "priority": "S"},
                    {"code": "B", "action": "del", "priority": "S"},
                ],
                "results": [{
                    # Anteriorité
                    "code": "A",
                    "date": datetime.datetime(1750, 1, 1),
                    "value": "24",
                    "anterior": True,
                }, {
                    # Résultat Accueil
                    "code": "A",
                    "date": datetime.datetime(1750, 1, 1),
                    "value": "48",
                }],
            }],
        }


    Exemple de message de query::

        {
            "context": "query",
            "samples": [{
                "type": "patient",
                "sample_id": "123456789",
            }],
        }

    Exemple de message de résultat patient::

        {
            "context": "result",
            "samples": [{
                "type": "patient",
                "sample_id": "123456789",
                "specimen": "Urine",
                "results": [{
                    "code": "A",
                    "date": datetime.datetime(2016, 1, 27, 14, 54),
                    "value": "42",
                    "flags": ["HH"],
                    "comments": ["Je lui donne pas 3 jours"],
                }],
            }],
        }

    Exemple de message de résultat QC::

        {
            "context": "result",
            "samples": [{
                "type": "control",
                "sample_id": "45633VB",
                "control_lot": "45633VB",
                "control_level": "3",
                "specimen": "Whole blood",
                "results": [{
                    "code": "A",
                    "value": "42",
                    "date": datetime.datetime(2016, 1, 27, 14, 54),
                }],
            }],
        }

    """

    VALIDATOR = jsonschema.Draft4Validator(
        schema={"type": "date-time"},
        types={
            "string": tuple(list(six.string_types) + [type(None)]),
            "date-time": (datetime.datetime, type(None)),
        },
    )
    ELSE = "_else_"
    IGNORE = "_ignore_"

    class ValidationError(Exception):
        pass

    def __init__(self, samples=None, **kwargs):
        super(Worklist, self).__init__(samples=(samples or []), **kwargs)

    def get_worklist(self, unique_together=[], **kwargs):
        """Permet de découper un sample plusieurs samples

        En fonction des valeurs d'attributs sur les `orders`.

        Utile pour les drivers non ASTM (où le patient est mis au
        même niveau que le sample).

        .. code-block:: python

            for sample, action in worklist.get_worklist(
                unique_together=["action"],
                action={"del": worklist.IGNORE, worklist.ELSE: "ADD"},
            ):
                # [...]
        """
        samples = []
        for sample in self.samples:
            for couple, orders in self.get_orders(
                sample["orders"], unique_together, **kwargs
            ):
                new_sample = Sample(sample.copy())
                new_sample["orders"] = orders
                samples.append((new_sample,) + couple)
        return samples

    def get_bruker_worklist(self, **kwargs):
        """Permet de regrouper les samples sous un identifiant de plaque"""
        samples = []
        for sample in self.samples:
            new_sample = Sample(sample.copy())
            samples.append(new_sample)
        return samples

    def get_astm_worklist(self, unique_together=[], samples="orders", **kwargs):
        """Permet de découper les sample et de regrouper les patients

        Les patients sont groupés par l'attribut `patient_id`.

        .. code-block:: python

            for patient in worklist.get_astm_worklist(
                unique_together=["action", "priority"],
                action={"del": "C", worklist.ELSE: "N"},
                priority={"S": "S", worklist.ELSE: "R"},
            ):
                # Ici on remplis un segment P
                for sample, action, priority in patient.samples:
                    # Ici on remplis un segment O
        """
        patients = OrderedDict()
        for sample in self.samples:
            patients.setdefault(
                sample["patient_id"],
                Patient(
                    dict([(k, sample[k]) for k in PATIENT_ATTRIBUTES if k in sample])
                ),
            )
            patients[sample["patient_id"]].setdefault("samples", [])
            orphan_results = sample.results[:]
            for couple, orders in self.get_orders(
                sample[samples], unique_together, **kwargs
            ):
                new_sample = Sample(sample.copy())
                new_sample[samples] = orders
                order_codes = [o.code for o in orders]
                new_sample["results"] = []
                for r in sample.results:
                    if r["code"] in order_codes:
                        orphan_results.remove(r)
                        new_sample["results"].append(Result(r))
                patients[sample["patient_id"]]["samples"].append(
                    ((new_sample,) + couple)
                )
            if orphan_results and self.context != "result":
                for couple in patients[sample["patient_id"]]["samples"]:
                    if couple[0].sample_id == sample.sample_id:
                        couple[0].results.extend([Result(r) for r in orphan_results])

            if not sample.orders and self.context == "result":
                new_sample = Sample(sample.copy())
                couple = []
                for k in unique_together:
                    v = kwargs.get(k, {})
                    if k not in new_sample:
                        couple.append(None)
                        continue
                    key = v.get(new_sample[k], v.get(Worklist.ELSE, new_sample[k]))
                    couple.append(key)
                patients[sample["patient_id"]]["samples"].append(
                    ((new_sample,) + tuple(couple))
                )
        return [p for p in patients.values() if p.samples]

    @staticmethod
    def get_orders(orders, unique_together=[], **kwargs):
        r = OrderedDict()
        # key => [values to ignore]
        ignore = {}
        for key, values in kwargs.items():
            for k, v in values.items():
                if v == Worklist.IGNORE:
                    ignore.setdefault(key, []).append(k)
        for order in orders:
            for k, values in ignore.items():
                if order[k] in values:
                    break
            else:
                couple = []
                for k in unique_together:
                    v = kwargs.get(k, {})
                    key = v.get(order[k], v.get(Worklist.ELSE, order[k]))
                    couple.append(key)
                r.setdefault(tuple(couple), []).append(Order(order))
        return r.items()

    def get_or_create(self, **kwargs):
        """Permet d'ajouter un sample dans une worklist sans créer de doublon

        >>> worklist = Worklist(context="result")
        >>> sample = worklist.get_or_create(type="patient", sample_id="123456789")
        """
        return self.samples.get_or_create(**kwargs)

    def validate(self):
        """Valide la worklist avec jsonschema"""
        try:
            return self.VALIDATOR.validate(self, SCHEMA)
        except jsonschema.exceptions.ValidationError as error:
            raise self.ValidationError(error)

    def dumps(self):
        """Dumpe la Worklist en json (gestion du type datetime)"""
        return json.dumps(self, cls=utils.JSONEncoder)

    @classmethod
    def loads(cls, data):
        """Crée une Worklist à partir d'une chaine json"""
        doc = json.loads(data)
        return cls.from_json(doc)

    @classmethod
    def from_json(cls, doc):
        for sample in doc.get("samples", []):
            for k, v in sample.items():
                if k in (
                    "request_date",
                    "admission_date",
                    "birthdate",
                    "collection_date",
                    "collection_end_date",
                ):
                    sample[k] = utils.decode_isoformat(v)
            for result in sample.get("results", []):
                for rk, rv in result.items():
                    if rk in ("date",):
                        result[rk] = utils.decode_isoformat(rv)
                for lot in result.get("lots", []):
                    for lk, lv in lot.items():
                        if lk == "expiration":
                            lot[lk] = utils.decode_isoformat(lv)

        return cls(**doc)
